struct ProjGradAlgorithm{P} <: Algorithm{P}
    problem::P

    "gradient step size"
    τ::Float64
    function ProjGradAlgorithm(problem; τ)
        return new{typeof(problem)}(problem, τ)
    end
end

ProjGradAlgorithm(problem::DualTVL1ROFOpProblem) =
    ProjGradAlgorithm(problem, τ=inv(8 * normB(problem)))

struct ProjGradState{A,Tp,Wv,R,S,K1,K2}
    algorithm::A
    "dual optimization variable"
    p::Tp

    "scalar temporary 1"
    rv::Wv
    "scalar temporary 2"
    sv::Wv

    "matrix view on rv"
    r::R
    "matrix view on sv"
    s::S

    k1::K1
    k2::K2
end

function init(alg::ProjGradAlgorithm{<:DualTVL1ROFOpProblem})
    g = alg.problem.g
    λ = alg.problem.λ
    d = ndims(g)

    pv = zeros(d * length(reinterpret(Float64, g)))
    rv = zeros(eltype(g), length(g))
    sv = zero(rv)

    p = extend(reshape(reinterpret(SVector{d,eltype(g)}, pv), size(g)), StaticKernels.ExtensionNothing())
    r = reshape(rv, size(g))
    s = extend(reshape(sv, size(g)), StaticKernels.ExtensionReplicate())

    @inline kf1(pw) = @inbounds -divergence(pw) - g[pw.position]
    k1 = Kernel{ntuple(_->-1:1, d)}(kf1)

    @inline function kf2(pw, sw)
        @inbounds q = p[sw.position] - alg.τ * gradient(sw)
        return @inbounds q ./ max.(projnorm(q, false) ./ λ[sw.position], 1)
    end
    k2 = Kernel{ntuple(_->0:1, d)}(kf2)

    return ProjGradState(alg, p, rv, sv, r, s, k1, k2)
end

function step!(ctx::ProjGradState)
    alg = ctx.algorithm
    # r = Λ*p - g
    map!(ctx.k1, ctx.r, ctx.p)
    # s = B * r
    mul!(ctx.sv, alg.problem.B, ctx.rv)
    # p = proj(p - λΛ's)
    map!(ctx.k2, ctx.p, ctx.p, ctx.s)

    return ctx
end

fetch(ctx::ProjGradState) = ctx.p
